/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is JSIRC Library
 *
 * The Initial Developer of the Original Code is New Dimensions Consulting,
 * Inc. Portions created by New Dimensions Consulting, Inc. are
 * Copyright (C) 1999 New Dimenstions Consulting, Inc. All
 * Rights Reserved.
 *
 * Contributor(s):
 *  Robert Ginda, rginda@ndcico.com, original author
 *  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
 *  Jeremie Miller, jer@jeremie.com, generalized into js socket lib (for Jabber)
 *

    Modified:
        Sep-2003  sdk updated to working (more) state.
              using: chatzilla/connection-xpcom.js
        Dec-2004  sdk fixed outstanding issues
            0) functions are now in alphabetical order with the exception of:
                     async.
            1) open no longer returns any value.
            2) isAlive has been created for testing the connection.
            3) isOpen has been created for testing open/close status
            4) init has been created for ease of subclassing
            5) available will set isConnected false when necessary.
            6) read and write have been reworked for cleaner operation. 
   
    Attributes:
        nsISocketTransport.openInputStream arguments:
        socket.openInputFlags = 0;            open flags      
        socket.openInputSegmentSize = 0;      segmentSize
        socket.openInputSegmentCount = 0;     segmentCount
      
        nsISocketTransport.openOutputStream arguments:
        socket.openOutputFlags = 0;         open flags
        socket.openOutputSegmentSize = 0;     segmentSize
        socket.openOutputSegmentCount = 0;      segmentCount
*/

if (typeof(JS_LIB_LOADED)=='boolean') {

/*
    function:  Socket
    purpose:  Constructor for an instance of Socket
    interface:  
        socket( [ classID ] )
    
    arguments:
        classID : classesByID class-identifier
    
    returns:
        not applicable
    
    Note:  by default, Socket uses the local socket-transport-service, however
    callers looking for a specific implementation can pass in the class
    identifier as an alternative.
*/
function Socket()
{
    if( arguments.length )
        this.init( arguments[ 0 ] );
    else
        this.init();
}

/*
    function: available
    purpose:
        number of bytes waiting to be read

    interface:
        available()

    returns:
        integer : number of bytes waiting to be read
*/
Socket.prototype.available = function ()
{
    if( !this.isAlive() )
        return 0;

    var bytesAvailable = 0;
    try {
        bytesAvailable = this._inputInterface.available();
    }
    catch(exception)
    {
        this.isConnected = false;
        this._exception = exception;
    }

    return bytesAvailable;
}

/*
    function:  close
    purpose:  closes the socket connection
    interface:
        close()

    returns:
        nothing.
*/
Socket.prototype.close = function ()
{
    if( !this.isOpen() ) return;

    this.isOpenFlag = false;
    this.isConnected = false;
    // calls to _inputStream.close() and _outputStream.close() didn't 
    // function.
    this._transport.close( 0 );
}

/*
    function: init
    purpose:  provides initialization for the Socket class.
    interface:  
        socket( [ classID ] )
    arguments:
        classID : classesByID class-identifier (see Socket for info)
    returns:
        nothing.
*/
Socket.prototype.init = function()
{
    this.isOpenFlag = false;
    this.isConnected = false;

    this.openInputFlags = 0;
    this.openInputSegmentSize = 0;
    this.openInputSegmentCount = 0;
    
    this.openOutputFlags = 0;
    this.openOutputSegmentSize = 0;
    this.openOutputSegmentCount = 0;
    
    var defaultContractID = "@mozilla.org/network/socket-transport-service;1";
    //  var defaultClassID = "{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"
    
    var socketServiceClass;
    switch( arguments.length )
    {
        case 0:
              socketServiceClass = C.classes[ defaultContractID ];
          break;
    
        case 1:
          socketServiceClass = C.classesByID[ arguments[ 0 ] ];
          break;
    
        default:
          throw( "Socket.init: unexpected arguments" );
          break;
    }
    if (!socketServiceClass)
        throw ("Socket constructor: Couldn't get socket service class.");
    
    var socketService = socketServiceClass.getService();
    if (!socketService)
        throw ("Socket constructor: Couldn't get socket service.");

    this._socketService = jslibQI(socketService, "nsISocketTransportService");
}

/*
    function: isAlive
    purpose:  tests the connection to see if it still works.
    interface:
        isAlive()
    arguments:
        none.
    returns:
        boolean : true if the connection still works.

    Note:
        This function is not accurate if invoked in the same javascript
        stack crawl as open().
*/
Socket.prototype.isAlive = function()
{
    this.isConnected = ( this.isOpen() ? this._transport.isAlive() : false );
    return this.isConnected;
}

/*
    function: isOpen
    purpose:  returns true if the socket has been opened (which is
        different from connect).  Multiple invocations of open will
        fail if each old connection is not closed first.
    interface:
        isOpen()
    arguments:
        none.
    returns:
        boolean : true if open() has been invoked without close.
*/
Socket.prototype.isOpen = function()
{
    return this.isOpenFlag;
}

//Added by Captain Caveman to get the proxy info object for this connection.
Socket.prototype.GetProxyInfo = function(host, port)
{
    var cls = Components.classes["@mozilla.org/network/protocol-proxy-service;1"];
    var pps = cls.getService(Components.interfaces.nsIProtocolProxyService);

    if (!pps)
        throw ("Couldn't get protocol proxy service");

    var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    var spec = 'pong:// ' + host + ':' + port;
    var uri = ios.newURI(spec,null,null);

    var info = null;
    if ("resolve" in pps)
        info = pps.resolve(uri, 0);
    else if ("examineForProxy" in pps)
        info = pps.examineForProxy(uri);

    return info;
}



/*
    function: open
    purpose:  opens the socket
    interface:  
        open( host, port [ binary ] )

    arguments:
        host : String, host to connect to
        port : int, port number to use
        binary : optional, use binary input support, defaults false
    
    returns:
        nothing.
*/
Socket.prototype.open = function (host, port, useproxy)
{
    if( this.isOpen() ) return;
    var openFlags = 0;
	var proxyinfo = null;

    this.host = host.toLowerCase();
    this.port = port;
    this.binary = ( arguments.length > 2 ) ? arguments[ 2 ] : false;
    this.isOpenFlag = true;

	if (useproxy == true) proxyinfo = this.GetProxyInfo(host, port);

    this._transport = this._socketService.createTransport ( null, 0, host, port, proxyinfo );
    if (!this._transport) throw ("Socket.open: Error opening transport.");

    this._inputStream = this._transport.openInputStream(this.openInputFlags, this.openInputSegmentSize, this.openInputSegmentCount);
    if (!this._inputStream) throw ("Socket.open: Error getting input stream.");
    if ( this.binary )
        this._inputInterface = this.toBinaryInputStream( this._inputStream );
    else
        this._inputInterface = this.toScriptableInputStream( this._inputStream );

    this._outputStream = this._transport.openOutputStream(this.openOutputFlags, this.openOutputSegmentSize, this.openOutputSegmentCount);
    if (!this._outputStream) throw ("Socket.open: Error getting output stream.");
	
    this.isConnected = true;
}

Socket.prototype.create = function (transport)
{
    if( this.isOpen() ) return;
    var openFlags = 0;

    this.isOpenFlag = true;	
	this.binary = true;
    this._transport = transport;

    this._inputStream = this._transport.openInputStream(this.openInputFlags, this.openInputSegmentSize, this.openInputSegmentCount);
    if (!this._inputStream) throw ("Socket.open: Error getting input stream.");
    if ( this.binary )
        this._inputInterface = this.toBinaryInputStream( this._inputStream );
    else
        this._inputInterface = this.toScriptableInputStream( this._inputStream );

    this._outputStream = this._transport.openOutputStream(this.openOutputFlags, this.openOutputSegmentSize, this.openOutputSegmentCount);
    if (!this._outputStream) throw ("Socket.open: Error getting output stream.");

	this.isConnected = true;
}

/*
    function: read
    purpose:
        reads data from a socket.

    interface:
        read( bytes )

    arguments:
        bytes : integer, number of bytes to read in

    returns:
        string

    Note:
        will only return the smaller of specified vs available bytes.

*/
Socket.prototype.read = function ( bytes )
{
    if ( !this.isAlive() )
        throw "Socket.read: Not Connected.";

    var returnValue = new String;
    if( bytes == 0 )
        return returnValue;

    var availableBytes = this.available();
    if( availableBytes == 0 )
        return returnValue;

    bytes = Math.min( availableBytes, bytes );
    if ( bytes )
        if ( this.binary )
            // despite the documentation, this call works
            returnValue = this._inputInterface.readBytes( bytes );
        else
            returnValue = this._inputInterface.read( bytes );
    
    return returnValue;
}

/*
    function: write
    purpose:
        writes the given string to the socket.

    interface:
        write( str )

    arguments:
        str : string to be written.

    returns:
        integer : number of bytes written.
*/
Socket.prototype.write = function (str)
{
    if( !this.isAlive() )
        throw "Socket.write: Not Connected.";

    var rv = 0;

    try
    {
        rv = this._outputStream.write(str, str.length);
    }
    catch (ex)
    {
        this.isConnected = false;
    }
    
    return rv;
}

/*
    function:  toBinaryInputStream
    purpose: creates an nsIBinaryInputStream wrapper around the given
    inputStream.

    interface:
          toBinaryInputStream( inputStream )

    arguments:
        inputStream : result of openInputStream

    returns:
        nsIBinaryInputStream
*/
Socket.prototype.toBinaryInputStream = function( inputStream )
{
    var si = jslibCreateInstance("@mozilla.org/binaryinputstream;1", 
            "nsIBinaryInputStream");
    si.setInputStream( inputStream );

    return si;
}

/*
    function: toScriptableInputStream
    purpose:
        creates an nsIScriptableInputStream wrapper around the given
        inputStream.

    interface:
        toScriptableInputStream( inputStream )

    arguments:
        inputStream : result of openInputStream

    returns:
        nsIScriptableInputStream
*/
Socket.prototype.toScriptableInputStream = function( inputStream )
{            
    var si = jslibCreateInstance("@mozilla.org/scriptableinputstream;1", 
                                 "nsIScriptableInputStream");
    si.init( inputStream );

    return si;
}

/*
    function: async
    purpose:
    interface:

    returns:
    usage:

     var aSocket = new Socket();
     var observer = {
        streamStarted: function(socketContext){ },           //onstart action
        streamStopped: function(socketContext, status){ },   //onstop action
        receiveData:   function(data){alert(data)}
      }
     aSocket.open("ftp.mozilla.org", 21);
     aSocket.async(observer );

*/
Socket.prototype.async = function (observer)
{
    // to preserve ourselves within necko/async
    this.wrappedJSObject = this;

    this._pump = jslibCreateInstance("@mozilla.org/network/input-stream-pump;1",
                                     "nsIInputStreamPump");

    this._pump.init(this._inputStream, -1, -1, 0, 0, false);
    this._pump.asyncRead(new SocketListener(observer), this);
}

// async callbacks
function SocketListener( observer )
{
    this._observer = observer;
}

SocketListener.prototype.onStartRequest = 
    function (channel, socketContext)
    {
        theSocket = socketContext.wrappedJSObject;
        this._observer.streamStarted( theSocket );
    }

SocketListener.prototype.onStopRequest = 
    function (channel, socketContext, status, errorMsg)
    {
        theSocket = socketContext.wrappedJSObject;
        this._observer.streamStopped( theSocket, status );
    }

SocketListener.prototype.onDataAvailable = 
    function (channel, socketContext, inputStream, sourceOffset, count)
    {
        theSocket = socketContext.wrappedJSObject;
        // try and maintain the connection
        // but read here because Socket.read will fail with HTTP requests due
        // to the socket being closed.
		if (typeof(JS_LIB_LOADED) != 'boolean') return; //Lame protection (not jslib, but captain caveman's) against calls when jslib is already unloaded.

        theSocket._inputStream = inputStream;
        if( theSocket.binary )
          {
            theSocket._inputInterface =
                theSocket.toBinaryInputStream( theSocket._inputStream );
            // despite the documentation, readBytes works
            this._observer.receiveData( theSocket._inputInterface.readBytes(count) );
          }
        else
          {
            theSocket._inputInterface =
                theSocket.toScriptableInputStream( theSocket._inputStream );
            this._observer.receiveData( theSocket._inputInterface.read(count) );
          }
    }
};  // END socket.js
